package de.fhdw.wtf.common.ast;

import java.util.Iterator;
import java.util.List;

import de.fhdw.wtf.common.ast.type.Type;
import de.fhdw.wtf.common.ast.visitor.ConstructorReferenceStateVisitor;
import de.fhdw.wtf.common.ast.visitor.ConstructorReferenceStateVisitorException;
import de.fhdw.wtf.common.ast.visitor.ConstructorReferenceStateVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.ConstructorReferenceStateVisitorReturnException;
import de.fhdw.wtf.common.constants.parser.AstDescriptionConstants;

/**
 * This class represents a state of a {@link ConstructorReference}, in which a constructor is referenced by the type and
 * the paramsTypes. After the referencer has done his work, the state will change into an
 * {@link ConstructorInvalidState} or a {@link ConstructorByReferenceState}.
 */
public final class ConstructorByTypeAndSignatureState implements ConstructorReferenceState {
	
	/**
	 * The type of the possible target.
	 */
	private final Type type;
	
	/**
	 * The target-ParamTypes.
	 */
	private final List<Type> paramTypes;
	
	/**
	 * Constructor of {@link ConstructorByTypeAndSignatureState}.
	 * 
	 * @param type
	 *            type
	 * @param paramTypes
	 *            paramTypes
	 */
	private ConstructorByTypeAndSignatureState(final Type type, final List<Type> paramTypes) {
		this.type = type;
		this.paramTypes = paramTypes;
	}
	
	/**
	 * Creates a {@link ConstructorByTypeAndSignatureState}-Object.
	 * 
	 * @param type
	 *            type
	 * @param paramTypes
	 *            paramTypes
	 * @return The ConstructorByNameState-Object
	 */
	public static ConstructorByTypeAndSignatureState create(final Type type, final List<Type> paramTypes) {
		return new ConstructorByTypeAndSignatureState(type, paramTypes);
	}
	
	/**
	 * Get the Type of the target.
	 * 
	 * @return type
	 */
	public Type getType() {
		return this.type;
	}
	
	/**
	 * Get the List of params of the target-Constructor.
	 * 
	 * @return name
	 */
	public List<Type> getParamTypes() {
		return this.paramTypes;
	}
	
	@Override
	public boolean equals(final Object o) {
		if (o instanceof ConstructorByTypeAndSignatureState) {
			final ConstructorByTypeAndSignatureState other = (ConstructorByTypeAndSignatureState) o;
			return this.getType().equals(other.getType()) && this.getParamTypes().equals(other.getParamTypes());
		}
		return false;
	}
	
	@Override
	public int hashCode() {
		return this.getType().hashCode() ^ this.getParamTypes().hashCode();
	}
	
	@Override
	public String toString() {
		final StringBuilder result = new StringBuilder();
		result.append(this.getType().getTypeString());
		result.append(AstDescriptionConstants.BRACKET_OPEN_TOKEN);
		final Iterator<Type> iterator = this.getParamTypes().iterator();
		while (iterator.hasNext()) {
			final Type current = iterator.next();
			result.append(current.getTypeString());
			if (iterator.hasNext()) {
				result.append(AstDescriptionConstants.COMMA_TOKEN);
			}
		}
		result.append(AstDescriptionConstants.BRACKET_CLOSE_TOKEN);
		return result.toString();
	}
	
	@Override
	public void accept(final ConstructorReferenceStateVisitor v) {
		v.handle(this);
	}
	
	@Override
	public <X> X accept(final ConstructorReferenceStateVisitorReturn<X> v) {
		return v.handle(this);
	}
	
	@Override
	public <Y extends Exception> void accept(final ConstructorReferenceStateVisitorException<Y> v) throws Y {
		v.handle(this);
	}
	
	@Override
	public <X, Y extends Exception> X accept(final ConstructorReferenceStateVisitorReturnException<X, Y> v) throws Y {
		return v.handle(this);
	}
}
