package de.fhdw.wtf.common.ast.type;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import de.fhdw.wtf.common.ast.visitor.AtomicTypeVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitor;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitorException;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitorReturnException;
import de.fhdw.wtf.common.ast.visitor.TypeVisitorReturn;
import de.fhdw.wtf.common.constants.parser.AstDescriptionConstants;
import de.fhdw.wtf.common.exception.parser.SumIsAnythingException;
import de.fhdw.wtf.common.token.Token;

/**
 * A {@link ProductType} is an anonymous class, which is composed by {@link ProductElementType}s.
 */
public final class ProductType extends CompositeType {
	
	/**
	 * generated.
	 */
	private static final long serialVersionUID = -6722262002471629663L;
	
	/**
	 * Abstract-Prototype is also a standardized Type of this, but the attributes have also standardized names!
	 */
	private ProductType abstractPrototype;
	
	/**
	 * Elements of this product.
	 */
	private final List<ProductElementType> elements;
	
	/**
	 * Constructor of {@link ProductType}.
	 * 
	 * @param firstToken
	 *            firstToken
	 * @param lastToken
	 *            lastToken
	 */
	private ProductType(final Token firstToken, final Token lastToken) {
		super(firstToken, lastToken);
		this.elements = new ArrayList<>();
	}
	
	/**
	 * Creates a {@link ProductType}-Object.
	 * 
	 * @param firstToken
	 *            firstToken
	 * @param lastToken
	 *            lastToken
	 * @return a ProductType-Object
	 */
	public static ProductType create(final Token firstToken, final Token lastToken) {
		return new ProductType(firstToken, lastToken);
	}
	
	/**
	 * Creates a {@link ProductType}-Object.
	 * 
	 * @param firstToken
	 *            firstToken
	 * @return a ProductType-Object
	 */
	public static ProductType create(final Token firstToken) {
		return new ProductType(firstToken, null);
	}
	
	/**
	 * Add a Element to the list of elements of this productType.
	 * 
	 * @param element
	 *            element to add
	 */
	public void addElement(final ProductElementType element) {
		this.getElements().add(element);
	}
	
	/**
	 * Returned the list of elements.
	 * 
	 * @return elements
	 */
	public List<ProductElementType> getElements() {
		return this.elements;
	}
	
	@Override
	public int compareTo(final Type o) {
		return o.accept(new TypeVisitorReturn<Integer>() {
			
			@Override
			public Integer handle(final AtomicType s) {
				return s.accept(new AtomicTypeVisitorReturn<Integer>() {
					
					@Override
					public Integer handle(final BaseType baseType) {
						return 1;
					}
					
					@Override
					public Integer handle(final ClassType clazz) {
						return -1;
					}
				});
			}
			
			@Override
			public Integer handle(final CompositeType c) {
				return c.accept(new CompositeTypeVisitorReturn<Integer>() {
					
					@Override
					public Integer handle(final ListType list) {
						return 1;
					}
					
					@Override
					public Integer handle(final MapType map) {
						return 1;
					}
					
					@Override
					public Integer handle(final ProductType product) {
						if (ProductType.this.getElements().size() > product.getElements().size()) {
							return 1;
						}
						if (ProductType.this.getElements().size() < product.getElements().size()) {
							return -1;
						}
						
						final List<ProductElementType> thisSorted = ProductType.this.sortElements();
						final List<ProductElementType> otherSorted = product.sortElements();
						
						for (int i = 0; i < thisSorted.size(); i++) {
							final int currentCompare = thisSorted.get(i).compareTo(otherSorted.get(i));
							if (currentCompare != 0) {
								return currentCompare;
							}
						}
						return 0;
					}
					
					@Override
					public Integer handle(final SumType sum) {
						return 1;
					}
					
					@Override
					public Integer handle(final ThrownType thrownType) {
						return 1;
					}
				});
			}
			
			@Override
			public Integer handle(final TypeProxy s) {
				return 1;
			}
		});
	}
	
	/**
	 * This method returns a sorted list of the elements of <code>this</code>.<br>
	 * 
	 * @return sorted elements
	 */
	public List<ProductElementType> sortElements() {
		final List<ProductElementType> types = new ArrayList<>();
		for (final ProductElementType element : this.getElements()) {
			types.add(element);
		}
		Collections.sort(types);
		return types;
	}
	
	@Override
	public boolean equals(final Object o) {
		if (o instanceof ProductType) {
			final ProductType other = (ProductType) o;
			if (this.getElements().size() != other.getElements().size()) {
				return false;
			}
			final List<ProductElementType> thisElements = this.getElements();
			final List<ProductElementType> otherElements = other.getElements();
			for (int i = 0; i < thisElements.size(); i++) {
				if (!thisElements.get(i).equals(otherElements.get(i))) {
					return false;
				}
			}
			return true;
		}
		return false;
	}
	
	@Override
	public int hashCode() {
		int result = 0;
		for (final ProductElementType type : this.getElements()) {
			result ^= type.hashCode();
		}
		return result;
	}
	
	@Override
	public String toString() {
		final StringBuilder result = new StringBuilder();
		result.append(AstDescriptionConstants.BRACKET_OPEN_TOKEN);
		final Iterator<ProductElementType> it = this.elements.iterator();
		while (it.hasNext()) {
			final ProductElementType current = it.next();
			result.append(current);
			if (it.hasNext()) {
				result.append(',');
			}
		}
		result.append(AstDescriptionConstants.BRACKET_CLOSE_TOKEN);
		return result.toString();
	}
	
	@Override
	public String getTypeString() {
		final StringBuilder result = new StringBuilder();
		result.append(AstDescriptionConstants.BRACKET_OPEN_TOKEN);
		final Iterator<ProductElementType> it = this.elements.iterator();
		while (it.hasNext()) {
			final ProductElementType current = it.next();
			result.append(current.getTypeString());
			if (it.hasNext()) {
				result.append(',');
			}
		}
		result.append(AstDescriptionConstants.BRACKET_CLOSE_TOKEN);
		return result.toString();
	}
	
	@Override
	public void accept(final CompositeTypeVisitor v) {
		v.handle(this);
	}
	
	@Override
	public <X> X accept(final CompositeTypeVisitorReturn<X> v) {
		return v.handle(this);
	}
	
	@Override
	public <X extends Exception> void accept(final CompositeTypeVisitorException<X> v) throws X {
		v.handle(this);
		
	}
	
	@Override
	public <X, Y extends Exception> X accept(final CompositeTypeVisitorReturnException<X, Y> v) throws Y {
		return v.handle(this);
	}
	
	/**
	 * If it set from the {@link de.fhdw.wtf.walker.tasks.PrototypesTask}, then this returned a standardized
	 * represantation of this Type. A AbstractPrototyp returned itself as abstractPrototyp and prototype. An
	 * abstractPrototype has one different to an prototype: The names of attributes are also standardized.
	 * 
	 * @return abstractPrototype
	 */
	public ProductType getAbstractPrototype() {
		return this.abstractPrototype;
	}
	
	/**
	 * Set the abstractPrototype.
	 * 
	 * @param abstractPrototype
	 *            abstractPrototype
	 */
	public void setAbstractPrototype(final ProductType abstractPrototype) {
		this.abstractPrototype = abstractPrototype;
	}
	
	@Override
	public boolean isSupertypeOf(final Type type) {
		return false;
	}
	
	@Override
	public ProductType getPrototype() {
		return (ProductType) super.getPrototype();
	}
	
	@Override
	public Type standardize() throws SumIsAnythingException {
		return this;
	}
	
	@Override
	public boolean isThrownType() {
		return false;
	}
}
