package de.fhdw.wtf.common.ast.type;

import java.util.LinkedList;
import java.util.List;

import de.fhdw.wtf.common.ast.DatabaseIDState;
import de.fhdw.wtf.common.ast.visitor.TypeProxyStateVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.TypeProxyStateVisitorReturnException;
import de.fhdw.wtf.common.ast.visitor.TypeVisitor;
import de.fhdw.wtf.common.ast.visitor.TypeVisitorException;
import de.fhdw.wtf.common.ast.visitor.TypeVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.TypeVisitorReturnException;
import de.fhdw.wtf.common.constants.walker.ExceptionConstants;
import de.fhdw.wtf.common.exception.ast.ASTException;
import de.fhdw.wtf.common.token.Token;

/**
 * This class represents a reference to a {@link AtomicType}. The way the {@link AtomicType} is referenced is defined by
 * the {@link TypeProxyState} of this {@link TypeProxy}.
 */
public class TypeProxy extends Type {
	
	/**
	 * generated.
	 */
	private static final long serialVersionUID = -6057509945574873081L;
	
	/**
	 * State of the TypeProxy.
	 */
	private TypeProxyState state;
	
	/**
	 * private TypeThrowingState throwingState;
	 * 
	 * /** Constructor of {@link TypeProxy}.
	 * 
	 * @param firstToken
	 *            firstToken
	 * @param state
	 *            state
	 * @param lastToken
	 *            lastToken
	 */
	protected TypeProxy(final Token firstToken, final TypeProxyState state, final Token lastToken) {
		super(firstToken, lastToken);
		this.state = state;
	}
	
	/**
	 * Creates a {@link TypeProxy}-Object.
	 * 
	 * @param firstToken
	 *            firstToken
	 * @param state
	 *            state
	 * @param lastToken
	 *            lastToken
	 * @return The TypeProxy-Object
	 */
	public static TypeProxy create(final Token firstToken, final TypeProxyState state, final Token lastToken) {
		return new TypeProxy(firstToken, state, lastToken);
	}
	
	/**
	 * Creates a {@link TypeProxy}-Object.
	 * 
	 * @param firstToken
	 *            firstToken
	 * @param state
	 *            states
	 * @return The TypeProxy-Object
	 */
	public static TypeProxy create(final Token firstToken, final TypeProxyState state) {
		return new TypeProxy(firstToken, state, null);
	}
	
	/**
	 * Changes the state of this to <code>state</code>.
	 * 
	 * @param newState
	 *            - new State for {@link TypeProxy}
	 */
	public void changeState(final TypeProxyState newState) {
		this.setState(newState);
	}
	
	// ///// PROJECTIONS ////////
	
	/**
	 * Returned the State of the TypeProxy.
	 * 
	 * @return State
	 */
	public TypeProxyState getState() {
		return this.state;
	}
	
	/**
	 * Set the State of the TypeProxy.
	 * 
	 * @param state
	 *            State
	 */
	private void setState(final TypeProxyState state) {
		this.state = state;
	}
	
	@Override
	public int compareTo(final Type o) {
		return o.accept(new TypeVisitorReturn<Integer>() {
			
			@Override
			public Integer handle(final AtomicType s) {
				return -1;
			}
			
			@Override
			public Integer handle(final CompositeType c) {
				return -1;
			}
			
			@Override
			public Integer handle(final TypeProxy s) {
				return s.getState().compareTo(TypeProxy.this.getState());
			}
		});
	}
	
	@Override
	public boolean isSupertypeOf(final Type type) {
		return this.getState().accept(new TypeProxyStateVisitorReturn<Boolean>() {
			
			@Override
			public Boolean handle(final ByNameState typeReferenceByName) {
				return false;
			}
			
			@Override
			public Boolean handle(final InvalidState invalidTypeReference) {
				return false;
			}
			
			@Override
			public Boolean handle(final ByReferenceState byType) {
				return byType.getTarget().isSupertypeOf(type);
			}
		});
	}
	
	@Override
	public Type getPrototype() {
		return super.getPrototype();
	}
	
	@Override
	public void setPrototype(final Type prototype) {
		if (!(prototype instanceof TypeProxy)) {
			throw new ASTException("The prototype \"" + prototype.getClass() + "\" does not match this type \""
					+ this.getClass() + "\"!");
		}
		super.setPrototype(prototype);
	}
	
	@Override
	public DatabaseIDState getTypeId() {
		return this.state.getTypeId();
	}
	
	@Override
	public void setTypeId(final DatabaseIDState typeId) {
		this.state.setTypeId(typeId);
	}
	
	/**
	 * Returns true if this and type are equals or if this state is the ByReferenceState and the target type is equals
	 * type. Returns false otherwise.
	 * 
	 * @param type
	 *            the type
	 * @return a boolean indicating what is said above.
	 */
	public boolean equalsOrTargetEquals(final Type type) {
		return this.equals(type) || this.state.accept(new TypeProxyStateVisitorReturn<Boolean>() {
			
			@Override
			public Boolean handle(final ByNameState byName) {
				return false;
			}
			
			@Override
			public Boolean handle(final InvalidState invalid) {
				return false;
			}
			
			@Override
			public Boolean handle(final ByReferenceState byReference) {
				return type.equals(byReference.getTarget());
			}
		});
	}
	
	/**
	 * Returns the target of this in case this is in {@link ByReferenceState}; throws an exception if not.
	 * 
	 * @return the target type of this.
	 * @throws ASTException
	 *             when not in {@link ByReferenceState}.
	 */
	public Type getTarget() throws ASTException {
		return this.state.accept(new TypeProxyStateVisitorReturnException<Type, ASTException>() {
			
			@Override
			public Type handle(final ByNameState byName) throws ASTException {
				throw new ASTException(ExceptionConstants.NO_TYPE_PROXY_TARGET_AVAILABLE);
			}
			
			@Override
			public Type handle(final InvalidState invalid) throws ASTException {
				throw new ASTException(ExceptionConstants.NO_TYPE_PROXY_TARGET_AVAILABLE);
			}
			
			@Override
			public Type handle(final ByReferenceState byReference) throws ASTException {
				return byReference.getTarget();
			}
		});
	}
	
	@Override
	public boolean equals(final Object obj) {
		if (obj instanceof TypeProxy) {
			final TypeProxy o = (TypeProxy) obj;
			return TypeProxy.this.getState().equals(o.getState());
		} else {
			return this.state.equals(obj);
		}
	}
	
	@Override
	public List<Type> getSuperTypes() {
		return this.state.accept(new TypeProxyStateVisitorReturn<List<Type>>() {
			
			@Override
			public List<Type> handle(final ByNameState byName) {
				return new LinkedList<>();
			}
			
			@Override
			public List<Type> handle(final InvalidState invalid) {
				return new LinkedList<>();
			}
			
			@Override
			public List<Type> handle(final ByReferenceState byReference) {
				return byReference.getTarget().getSuperTypes();
			}
			
		});
	}
	
	@Override
	public int hashCode() {
		return this.getState().hashCode();
	}
	
	@Override
	public String toString() {
		return this.getState().toString();
		// return TypeProxy.class.getSimpleName().toString() + " with state: " +
		// this.getState().toString();
	}
	
	@Override
	public String getTypeString() {
		return this.getState().getTypeString();
	}
	
	@Override
	public void accept(final TypeVisitor v) {
		v.handle(this);
	}
	
	@Override
	public <X> X accept(final TypeVisitorReturn<X> v) {
		return v.handle(this);
	}
	
	@Override
	public <X extends Exception> void accept(final TypeVisitorException<X> v) throws X {
		v.handle(this);
	}
	
	@Override
	public <X, Y extends Exception> X accept(final TypeVisitorReturnException<X, Y> v) throws Y {
		return v.handle(this);
	}
	
	@Override
	public Type standardize() {
		return this;
	}
	
	@Override
	public boolean isThrownType() {
		return false;
	}
}
