package de.fhdw.wtf.common.token;

import java.io.Serializable;

/**
 * A token represents a scanned part of the input grammar. Each token-type defines a regular sequence of characters
 * belonging to its type.
 * 
 * This type is the interface describing all tokens.
 */
public abstract class Token implements Serializable {
	
	/**
	 * generated.
	 */
	private static final long serialVersionUID = 6143241438846758818L;
	
	/**
	 * The position is the begin position of this token in the original input document.
	 */
	private final Position position;
	
	/**
	 * Constructor initializing the position of this token.
	 * 
	 * @param position
	 *            begin position of this token
	 */
	protected Token(final Position position) {
		this.position = position;
	}
	
	/**
	 * @return Returns true, if and only if this token represents any keyword.
	 */
	public boolean isKeywordToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents keyword: "class".
	 */
	public boolean isClassToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents keyword: "group".
	 */
	public boolean isGroupToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents any class modifier.
	 */
	public boolean isClassModifierToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents keyword: "function".
	 */
	public boolean isOperationToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents any operation modifier.
	 */
	public boolean isOperationModifierToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents operation modifier: "abstract".
	 */
	public boolean isAbstractOperationToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents class modifier: "abstract".
	 */
	public boolean isAbstractModifierToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents any attribute modifier.
	 */
	public boolean isAttributeModifierToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents attribute modifier: "prior".
	 */
	public boolean isPriorModifierToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents a comment.
	 */
	public boolean isCommentToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents an invalid token. The scanner uses this token to
	 *         denote unrecognized characters or unfinished comments.
	 */
	public boolean isInvalidToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents an identifier.
	 */
	public boolean isIdentifierToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents the end of a stream. This token always is the last
	 *         token in a stream.
	 */
	public boolean isEndToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents any symbol.
	 */
	public boolean isSymbolToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents symbol: "*".
	 */
	public boolean isAsteriskToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents symbol: ";".
	 */
	public boolean isSemicolonToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return @return Returns true, if and only if this token represents symbol: "{".
	 */
	public boolean isCurlyBracketOpenToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents symbol: "}".
	 */
	public boolean isCurlyBracketCloseToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents symbol: "+".
	 */
	public boolean isPlusSymbolToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents symbol: ":".
	 */
	public boolean isColonToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents symbol: ">".
	 */
	public boolean isGreaterSymbolToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents symbol: "=".
	 */
	public boolean isEqualToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents symbol: "[".
	 */
	public boolean isSquareBracketOpenToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents symbol: "]".
	 */
	public boolean isSquareBracketCloseToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents whitespace.
	 */
	public boolean isWhitespaceToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents an exception (keyword: "exception").
	 */
	public boolean isExceptionToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents a FindableToken.
	 */
	public boolean isFindableToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents a TransientToken.
	 */
	public boolean isTransientToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents a ServiceToken.
	 */
	public boolean isServiceToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents a HyphenToken.
	 */
	public boolean isHyphenToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents a PipeToken.
	 */
	public boolean isPipeToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents a BracketOpenToken.
	 */
	public boolean isBracketOpenToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents a BracketCloseToken.
	 */
	public boolean isBracketCloseToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents a CommaToken.
	 */
	public boolean isCommaToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents a MutableToken.
	 */
	public boolean isMutableToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents a VisitableToken.
	 */
	public boolean isVisitableToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents a SymmetricToken.
	 */
	public boolean isSymmetricToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents a DoubleSquareBracketOpenToken.
	 */
	public boolean isDoubleSquareBracketOpenToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents a DoubleSquareBracketCloseToken.
	 */
	public boolean isDoubleSquareBracketCloseToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents an ArrowToken.
	 */
	public boolean isArrowToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns true, if and only if this token represents an ExclamationToken.
	 */
	public boolean isExclamationToken() {
		return false; // must be overwritten as implementation
	}
	
	/**
	 * @return Returns the Tokenrepresentation length, <br>
	 *         e.g. A Token with the representation of "abc" returns 3.
	 */
	public abstract int getLength();
	
	@Override
	public boolean equals(final Object o) {
		if (o instanceof Token) {
			final Token t = (Token) o;
			return t.getPosition().equals(this.getPosition()) && this.equalsTemplate(t);
		}
		return false;
	}
	
	@Override
	public int hashCode() {
		return this.getPosition().hashCode() ^ this.hashCodeTemplate();
	}
	
	/**
	 * @param t
	 *            equal comparison element
	 * @return Returns true if and only if the token t is equal to this<br>
	 *         in terms of specialized attributes.
	 */
	protected abstract boolean equalsTemplate(Token t);
	
	/**
	 * Returns an hashCode-Value that depents on the equalsTemplate.
	 * 
	 * @return hashCode
	 */
	protected abstract int hashCodeTemplate();
	
	/**
	 * Creates a string representation of this token the same way as it occurred in the original text.
	 * 
	 * @return short string representation
	 */
	public abstract String stringRepresentation();
	
	@Override
	/**
	 * Creates a string representation of this token, including the position.
	 */
	public String toString() {
		return this.getClass().getSimpleName() + "[" + this.stringRepresentation() + "]" + " at " + this.getPosition();
	}
	
	/**
	 * @return the position of this token.
	 */
	public Position getPosition() {
		return this.position;
	}
	
}
